@@ -7,7 +7,6 @@ from copy import deepcopy
 
 import networkx as nx
 import numpy as np
-import ray
 from sklearn.base import BaseEstimator
 from sklearn.utils.validation import check_array, check_is_fitted
 
@@ -15,24 +14,23 @@ from hiclass.ConstantClassifier import ConstantClassifier
 from hiclass.HierarchicalClassifier import HierarchicalClassifier
 
 
-@ray.remote
-def _parallel_fit(lcppn, node):
-    classifier = lcppn.hierarchy_.nodes[node]["classifier"]
-    # get children examples
-    X, y = lcppn._get_successors(node)
-    unique_y = np.unique(y)
-    if len(unique_y) == 1 and lcppn.replace_classifiers:
-        classifier = ConstantClassifier()
-    classifier.fit(X, y)
-    return classifier
-
-
 class LocalClassifierPerParentNode(BaseEstimator, HierarchicalClassifier):
     """
     Assign local classifiers to each parent node of the graph.
 
     A local classifier per parent node is a local hierarchical classifier that fits one multi-class classifier
     for each parent node of the class hierarchy.
+
+    Examples
+    --------
+    >>> from hiclass import LocalClassifierPerParentNode
+    >>> y = [['1', '1.1'], ['2', '2.1']]
+    >>> X = [[1, 2], [3, 4]]
+    >>> lcppn = LocalClassifierPerParentNode()
+    >>> lcppn.fit(X, y)
+    >>> lcppn.predict(X)
+    array([['1', '1.1'],
+       ['2', '2.1']])
     """
 
     def __init__(
@@ -102,8 +100,6 @@ class LocalClassifierPerParentNode(BaseEstimator, HierarchicalClassifier):
 
         # TODO: Add parameter to receive hierarchy as parameter in constructor
 
-        # TODO: Add support to empty labels in some levels
-
         # Return the classifier
         return self
 
@@ -130,43 +126,38 @@ class LocalClassifierPerParentNode(BaseEstimator, HierarchicalClassifier):
         # Input validation
         X = check_array(X, accept_sparse="csr")
 
+        # Initialize array that holds predictions
         y = np.empty((X.shape[0], self.max_levels_), dtype=self.dtype_)
 
         # TODO: Add threshold to stop prediction halfway if need be
 
-        bfs = nx.bfs_successors(self.hierarchy_, source=self.root_)
-
         self.logger_.info("Predicting")
 
-        for predecessor, successors in bfs:
-            if predecessor == self.root_:
-                mask = [True] * X.shape[0]
-                subset_x = X[mask]
-            else:
-                mask = np.isin(y, predecessor).any(axis=1)
-                subset_x = X[mask]
-            if subset_x.shape[0] > 0:
-                classifier = self.hierarchy_.nodes[predecessor]["classifier"]
-                prediction = classifier.predict(subset_x)
-                level = nx.shortest_path_length(
-                    self.hierarchy_, self.root_, predecessor
-                )
-                if prediction.ndim == 2 and prediction.shape[1] == 1:
-                    prediction = prediction.flatten()
-                y[mask, level] = prediction
+        # Predict first level
+        classifier = self.hierarchy_.nodes[self.root_]["classifier"]
+        y[:, 0] = classifier.predict(X).flatten()
 
-        # Convert back to 1D if there is only 1 column to pass all sklearn's checks
-        if self.max_levels_ == 1:
-            y = y.flatten()
+        self._predict_remaining_levels(X, y)
 
-        # Remove separator from predictions
-        if y.ndim == 2:
-            for i in range(y.shape[0]):
-                for j in range(1, y.shape[1]):
-                    y[i, j] = y[i, j].split(self.separator_)[-1]
+        y = self._convert_to_1d(y)
+
+        self._remove_separator(y)
 
         return y
 
+    def _predict_remaining_levels(self, X, y):
+        for level in range(1, y.shape[1]):
+            predecessors = set(y[:, level - 1])
+            predecessors.discard("")
+            for predecessor in predecessors:
+                mask = np.isin(y[:, level - 1], predecessor)
+                predecessor_x = X[mask]
+                if predecessor_x.shape[0] > 0:
+                    successors = list(self.hierarchy_.successors(predecessor))
+                    if len(successors) > 0:
+                        classifier = self.hierarchy_.nodes[predecessor]["classifier"]
+                        y[mask, level] = classifier.predict(predecessor_x).flatten()
+
     def _initialize_local_classifiers(self):
         super()._initialize_local_classifiers()
         local_classifiers = {}
@@ -197,33 +188,18 @@ class LocalClassifierPerParentNode(BaseEstimator, HierarchicalClassifier):
         y = np.array(y)
         return X, y
 
-    def _fit_digraph_parallel(self, local_mode: bool = False):
-        self.logger_.info("Fitting local classifiers")
-        ray.init(num_cpus=self.n_jobs, local_mode=local_mode, ignore_reinit_error=True)
-        nodes = self._get_parents()
-        lcppn = ray.put(self)
-        results = [_parallel_fit.remote(lcppn, node) for node in nodes]
-        classifiers = ray.get(results)
-        for classifier, node in zip(classifiers, nodes):
-            self.hierarchy_.nodes[node]["classifier"] = classifier
+    @staticmethod
+    def _fit_classifier(self, node):
+        classifier = self.hierarchy_.nodes[node]["classifier"]
+        # get children examples
+        X, y = self._get_successors(node)
+        unique_y = np.unique(y)
+        if len(unique_y) == 1 and self.replace_classifiers:
+            classifier = ConstantClassifier()
+        classifier.fit(X, y)
+        return classifier
 
-    def _fit_digraph(self):
+    def _fit_digraph(self, local_mode: bool = False):
         self.logger_.info("Fitting local classifiers")
         nodes = self._get_parents()
-        for index, node in enumerate(nodes):
-            node_name = str(node).split(self.separator_)[-1]
-            self.logger_.info(
-                f"Fitting local classifier for node '{node_name}' ({index + 1}/{len(nodes)})"
-            )
-            classifier = self.hierarchy_.nodes[node]["classifier"]
-            # get children examples
-            X, y = self._get_successors(node)
-            unique_y = np.unique(y)
-            if len(unique_y) == 1 and self.replace_classifiers:
-                node_name = str(node).split(self.separator_)[-1]
-                self.logger_.warning(
-                    f"Fitting ConstantClassifier for node '{node_name}'"
-                )
-                self.hierarchy_.nodes[node]["classifier"] = ConstantClassifier()
-                classifier = self.hierarchy_.nodes[node]["classifier"]
-            classifier.fit(X, y)
+        self._fit_node_classifier(nodes, local_mode)
